@@ -1,5 +1,5 @@
 //////////////////////////////////////////////////////////////////////////////
-//    Copyright 2004-2014, SenseGraphics AB
+//    Copyright 2004-2019, SenseGraphics AB
 //
 //    This file is part of H3DUtil.
 //
@@ -29,10 +29,13 @@
 //////////////////////////////////////////////////////////////////////////////
 
 #include <H3DUtil/Image.h>
-#ifdef WIN32
+#ifdef H3D_WINDOWS
 #undef max
 #endif
 #include <limits>
+#ifdef HAVE_OPENEXR
+#include <OpenEXR/half.h>
+#endif
 
 using namespace H3DUtil;
 
@@ -150,16 +153,41 @@ namespace ImageInternals {
 
   inline H3DFloat getRationalValueAsFloat( void *i, 
                                            unsigned int bytes_to_read ) {
-    assert( bytes_to_read == 4 || bytes_to_read == 8 );
+    assert( bytes_to_read == 4 || bytes_to_read == 8||bytes_to_read==2 );
     double v = 0;
     if( bytes_to_read == 4 ) {
       v = *((float *)i);
     } else if( bytes_to_read == 8 ) {
       v = *((double *)i);
+    } else if(bytes_to_read == 2){ 
+#ifdef HAVE_OPENEXR
+      unsigned short v_temp;
+      memcpy( &v_temp, i, bytes_to_read ); 
+      half v_temp_half;
+      v_temp_half.setBits(v_temp);
+      v = float(v_temp_half);
+#else
+      Console(LogLevel::Error)<<"ERROR: need openexr support to convert 16bit float to 32bit float "<<std::endl;
+#endif
+
     }
     return (H3DFloat) v;
   }
 
+  inline H3DFloat getValueAsFloat( void *i, unsigned int bytes_to_read, 
+                                   Image::PixelComponentType pct ){
+    if( pct == Image::UNSIGNED ) {
+      return getUnsignedValueAsFloat(i, bytes_to_read);
+    }else if( pct == Image::SIGNED ) {
+      return getSignedValueAsFloat(i, bytes_to_read );
+    }else if( pct == Image::RATIONAL ){
+      return getRationalValueAsFloat(i, bytes_to_read);
+    }else{
+      Console(LogLevel::Error)<<"Warning: Specified pixel component type data converting to float is not supported yet"<<std::endl;
+    }
+    return 0;
+  }
+
   inline void writeFloatAsSignedValue( H3DFloat r,
                                        void *i, 
                                        unsigned int bytes_to_write ) {
@@ -183,22 +211,45 @@ namespace ImageInternals {
   inline void writeFloatAsRationalValue( H3DFloat r,
                                          void *i, 
                                          unsigned int bytes_to_write ) {
-    assert( bytes_to_write == 4 || bytes_to_write == 8 );
+    assert( bytes_to_write == 4 || bytes_to_write == 8 || bytes_to_write == 2);
 
     if( bytes_to_write == 4 ) {
       *((float *)i) = r;
     } else if( bytes_to_write == 8 ) {
       *((double *)i) = r;
+    } else if( bytes_to_write == 2 ) {
+#ifdef HAVE_OPENEXR
+      half r_temp_half = half(r);
+      unsigned short r_temp = r_temp_half.bits();
+      memcpy( i, &r_temp, bytes_to_write );
+#else
+      Console(LogLevel::Error)<<"ERROR: need openexr support to convert float to half"<<std::endl;
+#endif
+    }
+  }
+  inline void writeFloatAsValue( H3DFloat r,
+                                 void *i,
+                                 unsigned int bytes_to_write,
+                                 Image::PixelComponentType pct){
+    if( pct == Image::UNSIGNED ) {
+      writeFloatAsUnsignedValue( r, i, bytes_to_write );
+    }else if( pct == Image::SIGNED ) {
+      writeFloatAsSignedValue( r, i, bytes_to_write );
+    }else if( pct == Image::RATIONAL ) {
+      writeFloatAsRationalValue( r, i, bytes_to_write );
+    }else{
+      Console(LogLevel::Error)<<"Warning: Specified pixel component type data converting to pixel data is not supported yet"<<std::endl;
     }
   }
-
 }
 
 H3DUtil::RGBA Image::getPixel( int x, int y, int z ) {
   unsigned int byte_rem = bitsPerPixel() % 8;
   unsigned int bytes_per_pixel = bitsPerPixel() / 8;
 
-  assert( byte_rem == 0 );
+  if (byte_rem != 0) {
+    assert(!"bitsPerPixel value can not be divided by 8 for image");
+  }
   if( bytes_per_pixel <= 8 ) {
     char pixel_data[8]; 
     getElement( pixel_data, x, y, z );
@@ -217,7 +268,9 @@ void Image::setPixel( const H3DUtil::RGBA &value, int x, int y, int z ) {
   unsigned int byte_rem = bitsPerPixel() % 8;
   unsigned int bytes_per_pixel = bitsPerPixel() / 8;
 
-  assert( byte_rem == 0 );
+  if (byte_rem != 0) {
+    assert(!"bitsPerPixel value can not be divided by 8 for image");
+  }
   if( bytes_per_pixel <= 8 ) {
     char pixel_data[8];
     RGBAToImageValue( value, pixel_data );
@@ -230,6 +283,30 @@ void Image::setPixel( const H3DUtil::RGBA &value, int x, int y, int z ) {
   }
 }
 
+void Image::setAllPixels( const H3DUtil::RGBA &value ) {
+  assert( bitsPerPixel() % 8 == 0 );
+  unsigned int bytes_per_pixel = bitsPerPixel() / 8;
+  unsigned char* pixel_data = new unsigned char[bytes_per_pixel<=8? 8: bytes_per_pixel];
+  unsigned char* image_data = (unsigned char *)getImageData();
+  unsigned int w = width();
+  unsigned int h = height();
+  unsigned int d = depth();
+
+  RGBAToImageValue( value, pixel_data );
+
+  // copy every pixel into correct locations
+  for( unsigned int x = 0; x < w; ++x ) {
+    for( unsigned int y = 0; y < h; ++y ) {
+      for( unsigned int z = 0; z < d; ++z ) {
+        memcpy( &image_data[((z*h + y)*w + x)*bytes_per_pixel],
+          pixel_data,
+          bytes_per_pixel );
+      }
+    }
+  }
+  delete[] pixel_data;
+}
+
 H3DUtil::RGBA Image::imageValueToRGBA( void *_pixel_data ) {
   using namespace ImageInternals;
 
@@ -238,10 +315,12 @@ H3DUtil::RGBA Image::imageValueToRGBA( void *_pixel_data ) {
   unsigned int byte_rem = bitsPerPixel() % 8;
   unsigned int bytes_per_pixel = bitsPerPixel() / 8;
 
-  assert( byte_rem == 0 );
+  if( byte_rem!=0 ) {
+    assert(!"bitsPerPixel value can not be divided by 8 for image");
+  }
 
   switch( pixelType() ) {
-  case Image::LUMINANCE: 
+  case Image::LUMINANCE: {
     switch( pixelComponentType() ) { 
     case Image::UNSIGNED: {
       H3DFloat fv = getUnsignedValueAsFloat( pixel_data, 
@@ -253,12 +332,20 @@ H3DUtil::RGBA Image::imageValueToRGBA( void *_pixel_data ) {
                                            bytes_per_pixel );
       return H3DUtil::RGBA( fv, fv, fv, 1 );
     }
-    case Image::RATIONAL: {
+    case Image::RATIONAL:
+    case Image::RATIONAL_UNSIGNED: {
       H3DFloat fv = getRationalValueAsFloat( pixel_data, 
                                              bytes_per_pixel );
       return H3DUtil::RGBA( fv, fv, fv, 1 );
     }
-    };
+    }
+    break;
+  }
+  case Image::R:
+    {
+      H3DFloat fv = getValueAsFloat( pixel_data, bytes_per_pixel, pixelComponentType() );
+      return H3DUtil::RGBA(fv, 0,0,1);
+    }
   case Image::LUMINANCE_ALPHA: {
     unsigned int bytes_per_component = bytes_per_pixel / 2;
     switch( pixelComponentType() ) { 
@@ -277,14 +364,16 @@ H3DUtil::RGBA Image::imageValueToRGBA( void *_pixel_data ) {
 
       return H3DUtil::RGBA( fv, fv, fv, a );
     }
-    case Image::RATIONAL: {
+    case Image::RATIONAL:
+    case Image::RATIONAL_UNSIGNED: {
       H3DFloat fv = getRationalValueAsFloat( pixel_data, 
                                              bytes_per_component );
       H3DFloat a = getRationalValueAsFloat( pixel_data + bytes_per_component, 
                                             bytes_per_component );
       return H3DUtil::RGBA( fv, fv, fv, a );
     }
-    };
+    }
+    break;
   }
   case Image::RGB:
   case Image::VEC3: {
@@ -310,7 +399,8 @@ H3DUtil::RGBA Image::imageValueToRGBA( void *_pixel_data ) {
                                           bytes_per_component );
       return H3DUtil::RGBA( r, g, b, 1 );
     }
-    case Image::RATIONAL: {
+    case Image::RATIONAL:
+    case Image::RATIONAL_UNSIGNED: {
       H3DFloat r = getRationalValueAsFloat( pixel_data, 
                                             bytes_per_component );
       H3DFloat g = getRationalValueAsFloat( pixel_data + bytes_per_component, 
@@ -320,7 +410,8 @@ H3DUtil::RGBA Image::imageValueToRGBA( void *_pixel_data ) {
                                             bytes_per_component );
       return H3DUtil::RGBA( r, g, b, 1 );
     }
-    };
+    }
+    break;
   }
   case Image::BGR:  {
     unsigned int bytes_per_component = bytes_per_pixel / 3;
@@ -345,7 +436,8 @@ H3DUtil::RGBA Image::imageValueToRGBA( void *_pixel_data ) {
                                           bytes_per_component );
       return H3DUtil::RGBA( r, g, b, 1 );
     }
-    case Image::RATIONAL: {
+    case Image::RATIONAL:
+    case Image::RATIONAL_UNSIGNED: {
       H3DFloat b = getRationalValueAsFloat( pixel_data, 
                                             bytes_per_component );
       H3DFloat g = getRationalValueAsFloat( pixel_data + bytes_per_component, 
@@ -355,7 +447,8 @@ H3DUtil::RGBA Image::imageValueToRGBA( void *_pixel_data ) {
                                             bytes_per_component );
       return H3DUtil::RGBA( r, g, b, 1 );
     }
-    };
+    }
+    break;
   }
   case Image::RGBA:  {
     unsigned int bytes_per_component = bytes_per_pixel / 4;
@@ -386,7 +479,8 @@ H3DUtil::RGBA Image::imageValueToRGBA( void *_pixel_data ) {
                                           bytes_per_component );
       return H3DUtil::RGBA( r, g, b, a );
     }
-    case Image::RATIONAL: {
+    case Image::RATIONAL:
+    case Image::RATIONAL_UNSIGNED: {
       H3DFloat r = getRationalValueAsFloat( pixel_data, 
                                             bytes_per_component );
       H3DFloat g = getRationalValueAsFloat( pixel_data + bytes_per_component, 
@@ -399,7 +493,8 @@ H3DUtil::RGBA Image::imageValueToRGBA( void *_pixel_data ) {
                                             bytes_per_component );
       return H3DUtil::RGBA( r, g, b, a );
     }
-    };
+    }
+    break;
   }
   case Image::BGRA: {
     unsigned int bytes_per_component = bytes_per_pixel / 4;
@@ -430,7 +525,8 @@ H3DUtil::RGBA Image::imageValueToRGBA( void *_pixel_data ) {
                                           bytes_per_component );
       return H3DUtil::RGBA( r, g, b, a );
     }
-    case Image::RATIONAL: {
+    case Image::RATIONAL:
+    case Image::RATIONAL_UNSIGNED: {
       H3DFloat b = getRationalValueAsFloat( pixel_data, 
                                             bytes_per_component );
       H3DFloat g = getRationalValueAsFloat( pixel_data + bytes_per_component, 
@@ -443,9 +539,39 @@ H3DUtil::RGBA Image::imageValueToRGBA( void *_pixel_data ) {
                                             bytes_per_component );
       return H3DUtil::RGBA( r, g, b, a );
     }
-    };
+    }
+    break;
+  }
+  case Image::RG: {
+    unsigned int bytes_per_component = bytes_per_pixel / 2;
+    switch( pixelComponentType() ) {
+    case Image::UNSIGNED: {
+      H3DFloat r = getUnsignedValueAsFloat( pixel_data,
+        bytes_per_component );
+      H3DFloat g = getUnsignedValueAsFloat( pixel_data + bytes_per_component,
+        bytes_per_component );
+      return H3DUtil::RGBA( r, g, 0, 1 );
+    }
+    case Image::SIGNED: {
+      H3DFloat r = getSignedValueAsFloat( pixel_data,
+        bytes_per_component );
+      H3DFloat g = getSignedValueAsFloat( pixel_data + bytes_per_component,
+        bytes_per_component );
+      return H3DUtil::RGBA( r, g, 0, 1 );
+    }
+    case Image::RATIONAL:
+    case Image::RATIONAL_UNSIGNED: {
+      H3DFloat r = getRationalValueAsFloat( pixel_data,
+        bytes_per_component );
+      H3DFloat g = getRationalValueAsFloat( pixel_data + bytes_per_component,
+        bytes_per_component );
+      return H3DUtil::RGBA( r, g, 0, 1 );
+    }
+    }
+    break;
+  }
+  default: {}
   }
-  };
   return H3DUtil::RGBA();
 }
 
@@ -457,7 +583,9 @@ void Image::RGBAToImageValue( const H3DUtil::RGBA &rgba, void *_pixel_data ) {
   unsigned int byte_rem = bitsPerPixel() % 8;
   unsigned int bytes_per_pixel = bitsPerPixel() / 8;
 
-  assert( byte_rem == 0 );
+  if (byte_rem != 0) {
+    assert(!"bitsPerPixel value can not be divided by 8 for image");
+  }
 
   switch( pixelType() ) {
   case Image::LUMINANCE: 
@@ -474,13 +602,18 @@ void Image::RGBAToImageValue( const H3DUtil::RGBA &rgba, void *_pixel_data ) {
                                bytes_per_pixel );
       return;
     }
-    case Image::RATIONAL: {
+    case Image::RATIONAL:
+    case Image::RATIONAL_UNSIGNED: {
       writeFloatAsRationalValue( rgba.r,
                                  pixel_data, 
                                  bytes_per_pixel );
       return;
     }
     };
+  case Image::R:{
+    writeFloatAsValue(rgba.r, pixel_data, bytes_per_pixel, pixelComponentType() );
+    return;
+  }
   case Image::LUMINANCE_ALPHA: {
     unsigned int bytes_per_component = bytes_per_pixel / 2;
     switch( pixelComponentType() ) { 
@@ -502,7 +635,8 @@ void Image::RGBAToImageValue( const H3DUtil::RGBA &rgba, void *_pixel_data ) {
                                  bytes_per_component );
       return;
     }
-    case Image::RATIONAL: {
+    case Image::RATIONAL:
+    case Image::RATIONAL_UNSIGNED: {
       writeFloatAsRationalValue( rgba.r,
                                  pixel_data, 
                                  bytes_per_component );
@@ -511,7 +645,8 @@ void Image::RGBAToImageValue( const H3DUtil::RGBA &rgba, void *_pixel_data ) {
                                  bytes_per_component );
       return;
     }
-    };
+    }
+    break;
   }
   case Image::RGB:
   case Image::VEC3: {
@@ -541,7 +676,8 @@ void Image::RGBAToImageValue( const H3DUtil::RGBA &rgba, void *_pixel_data ) {
                                bytes_per_component );
       return;
     }
-    case Image::RATIONAL: {
+    case Image::RATIONAL:
+    case Image::RATIONAL_UNSIGNED: {
       writeFloatAsRationalValue( rgba.r,
                                pixel_data, 
                                bytes_per_component );
@@ -553,7 +689,8 @@ void Image::RGBAToImageValue( const H3DUtil::RGBA &rgba, void *_pixel_data ) {
                                bytes_per_component );
       return;
     }
-    };
+    }
+    break;
   }
   case Image::BGR:  {
     unsigned int bytes_per_component = bytes_per_pixel / 3;
@@ -582,7 +719,8 @@ void Image::RGBAToImageValue( const H3DUtil::RGBA &rgba, void *_pixel_data ) {
                                bytes_per_component );
       return;
     }
-    case Image::RATIONAL: {
+    case Image::RATIONAL:
+    case Image::RATIONAL_UNSIGNED: {
       writeFloatAsRationalValue( rgba.b,
                                  pixel_data, 
                                  bytes_per_component );
@@ -594,7 +732,8 @@ void Image::RGBAToImageValue( const H3DUtil::RGBA &rgba, void *_pixel_data ) {
                                  bytes_per_component );
       return;
     }
-    };
+    }
+    break;
   }
   case Image::RGBA:  {
     unsigned int bytes_per_component = bytes_per_pixel / 4;
@@ -629,7 +768,8 @@ void Image::RGBAToImageValue( const H3DUtil::RGBA &rgba, void *_pixel_data ) {
                                  bytes_per_component );
       return;
     }
-    case Image::RATIONAL: {
+    case Image::RATIONAL:
+    case Image::RATIONAL_UNSIGNED: {
       writeFloatAsRationalValue( rgba.r,
                                  pixel_data, 
                                  bytes_per_component );
@@ -645,7 +785,8 @@ void Image::RGBAToImageValue( const H3DUtil::RGBA &rgba, void *_pixel_data ) {
       return;
     }
 
-    };
+    }
+    break;
   }
   case Image::BGRA: {
    unsigned int bytes_per_component = bytes_per_pixel / 4;
@@ -680,7 +821,8 @@ void Image::RGBAToImageValue( const H3DUtil::RGBA &rgba, void *_pixel_data ) {
                                  bytes_per_component );
       return;
     }
-    case Image::RATIONAL: {
+    case Image::RATIONAL:
+    case Image::RATIONAL_UNSIGNED: {
       writeFloatAsRationalValue( rgba.b,
                                  pixel_data, 
                                  bytes_per_component );
@@ -696,14 +838,51 @@ void Image::RGBAToImageValue( const H3DUtil::RGBA &rgba, void *_pixel_data ) {
       return;
     }
 
-    };
+    }
+    break;
+  }
+  case Image::RG: {
+    unsigned int bytes_per_component = bytes_per_pixel / 2;
+    switch( pixelComponentType() ) {
+    case Image::UNSIGNED: {
+      writeFloatAsUnsignedValue( rgba.r,
+        pixel_data,
+        bytes_per_component );
+      writeFloatAsUnsignedValue( rgba.g,
+        pixel_data + bytes_per_component,
+        bytes_per_component );
+      return;
+    }
+    case Image::SIGNED: {
+      writeFloatAsSignedValue( rgba.r,
+        pixel_data,
+        bytes_per_component );
+      writeFloatAsSignedValue( rgba.g,
+        pixel_data + bytes_per_component,
+        bytes_per_component );
+      return;
+    }
+    case Image::RATIONAL:
+    case Image::RATIONAL_UNSIGNED: {
+      writeFloatAsRationalValue( rgba.r,
+        pixel_data,
+        bytes_per_component );
+      writeFloatAsRationalValue( rgba.g,
+        pixel_data + bytes_per_component,
+        bytes_per_component );
+      return;
+    }
+    }
+    break;
+  }
+  default: {}
   }
-  };
 }
 
 unsigned int Image::nrPixelComponents() {
   PixelType pixel_type = pixelType();
-  if( pixel_type == LUMINANCE ) return 1;
+  if( pixel_type == LUMINANCE ||
+      pixel_type == R ) return 1;
   if( pixel_type == LUMINANCE_ALPHA ) return 2;
   if( pixel_type == RGB ||
       pixel_type == BGR ||
@@ -724,7 +903,7 @@ namespace ImageInternals {
     
     for (unsigned int i = 0; i < nr_elements; ++i) {
       normalized_data[i] = 
-      (d[ i ] / FloatType( numeric_limits<A >::max() ) ) * scale + bias;
+      (d[ i ] / FloatType( std::numeric_limits<A >::max() ) ) * scale + bias;
       if( normalized_data[i] < 0 ) normalized_data[i] = 0;
     }
   }
@@ -747,8 +926,8 @@ namespace ImageInternals {
     FloatType *normalized_data = NULL;
     try {
       normalized_data = new FloatType[nr_voxels];
-    } catch (bad_alloc& ba) {
-      Console(4) << ba.what() << endl;
+    } catch (std::bad_alloc& ba) {
+      Console(LogLevel::Error) << ba.what() << std::endl;
       return NULL;
     }
     
