@@ -1,5 +1,5 @@
 //////////////////////////////////////////////////////////////////////////////
-//    Copyright 2004-2014, SenseGraphics AB
+//    Copyright 2004-2019, SenseGraphics AB
 //
 //    This file is part of H3DUtil.
 //
@@ -47,6 +47,10 @@ using namespace std;
 #include <mach/thread_act.h>
 #endif
 
+#ifdef HAVE_NVIDIATX
+#include <nvToolsExt.h>
+#endif
+
 #include <H3DUtil/Console.h>
 
 
@@ -58,6 +62,10 @@ using namespace std;
 #endif
 
 vector<ThreadBase *> ThreadBase::current_threads;
+
+MutexLock ThreadSpecificContainerBase::ThreadInfo::mutex;
+std::size_t ThreadSpecificContainerBase::ThreadInfo::next_thread_index = 0;
+
 // Constructor.
 MutexLock::MutexLock() {
   pthread_mutex_init( &mutex, NULL );
@@ -154,24 +162,24 @@ void ConditionLock::wait() {
 // certain time. If the time exceeds the specified time false
 // is returned. If signal is received true is returned.
 bool ConditionLock::timedWait( unsigned int millisecs ) {
-  int nanosecs, secs;
-  const int NANOSEC_PER_MILLISEC = (int)1e6;
-  const int NANOSEC_PER_SEC = (int)1e9;
+  long secs, nanosecs;
+  const long NANOSEC_PER_MILLISEC = (long)1e6;
+  const long NANOSEC_PER_SEC = (long)1e9;
 
   /* get current system time and add millisecs */
 
-#ifdef WIN32
+#ifdef H3D_WINDOWS
   _timeb sys_time;
   _ftime(&sys_time);
-  secs = sys_time.time;
-  nanosecs = ((int) (millisecs + sys_time.millitm)) * NANOSEC_PER_MILLISEC;
+  secs = (long)sys_time.time;
+  nanosecs = ( (millisecs + sys_time.millitm)) * NANOSEC_PER_MILLISEC;
 #else
   timeval sys_time;
   gettimeofday( &sys_time, NULL );
   secs = sys_time.tv_sec;
-  nanosecs = ((int) (millisecs + sys_time.tv_usec * 1e-3)) *
+  nanosecs = ((millisecs + sys_time.tv_usec * (long)1e-3)) *
     NANOSEC_PER_MILLISEC;
-#endif
+#endif // H3D_WINDOWS
   
   if (nanosecs >= NANOSEC_PER_SEC) {
       secs = secs + nanosecs / NANOSEC_PER_SEC;
@@ -179,8 +187,8 @@ bool ConditionLock::timedWait( unsigned int millisecs ) {
   } 
 
   timespec time;
-  time.tv_nsec = (long)nanosecs;
-  time.tv_sec = (long)secs;
+  time.tv_nsec = nanosecs;
+  time.tv_sec = secs;
 
   return pthread_cond_timedwait( &cond, &mutex, &time ) != ETIMEDOUT;
 }
@@ -210,11 +218,11 @@ void *PeriodicThread::thread_func( void * _data ) {
           mach_thread_self(),
           THREAD_TIME_CONSTRAINT_POLICY, (thread_policy_t)&ttcpolicy,
           THREAD_TIME_CONSTRAINT_POLICY_COUNT)) != KERN_SUCCESS ) {
-    H3DUtil::Console(4) << "Threads: set_realtime() failed" << endl;
+    H3DUtil::Console(LogLevel::Error) << "Threads: set_realtime() failed" << endl;
   }
 #endif
   
-#ifdef WIN32
+#ifdef H3D_WINDOWS
   HANDLE hTimer = NULL;
   LARGE_INTEGER liDueTime;
   // add some time to compensate for the time spend between waking up
@@ -228,14 +236,14 @@ void *PeriodicThread::thread_func( void * _data ) {
     // Create a waitable timer.
     hTimer = CreateWaitableTimer(NULL, TRUE, NULL );
     if (!hTimer) {
-      H3DUtil::Console(4) << "CreateWaitableTimer failed (%d)" << endl 
+      H3DUtil::Console(LogLevel::Error) << "CreateWaitableTimer failed (%d)" << endl 
                          << GetLastError() << endl;
       timeEndPeriod(1); 
       return NULL;
     }
   
     if (!SetWaitableTimer(hTimer, &liDueTime, 0, NULL, NULL, 0)) {
-      H3DUtil::Console(4) << "SetWaitableTimer failed (%d)\n"
+      H3DUtil::Console(LogLevel::Error) << "SetWaitableTimer failed (%d)\n"
         << GetLastError() << endl;
       timeEndPeriod(1); 
       return NULL;
@@ -244,7 +252,7 @@ void *PeriodicThread::thread_func( void * _data ) {
 
 #else
   TimeStamp last_time;
-#endif
+#endif // H3D_WINDOWS
 
   while( thread->thread_func_is_running ) {
 #ifdef THREAD_LOCK_DEBUG
@@ -255,13 +263,13 @@ void *PeriodicThread::thread_func( void * _data ) {
     }
 #endif
     if( thread->frequency > 0 ) {
-#ifdef WIN32
+#ifdef H3D_WINDOWS
       if (WaitForSingleObject(hTimer, INFINITE) != WAIT_OBJECT_0)
         printf("WaitForSingleObject failed (%d)\n", GetLastError());
 
       // Set a timer to wait for.
       if (!SetWaitableTimer(hTimer, &liDueTime, 0, NULL, NULL, 0)) {
-        H3DUtil::Console(4) << "SetWaitableTimer failed (%d)\n"
+        H3DUtil::Console(LogLevel::Error) << "SetWaitableTimer failed (%d)\n"
           << GetLastError() << endl;
         return NULL;
       }
@@ -274,7 +282,7 @@ void *PeriodicThread::thread_func( void * _data ) {
          usleep( 1e6 * delay );
       }
       last_time = TimeStamp();
-#endif
+#endif //H3D_WINDOWS
     }
 
 #ifdef THREAD_LOCK_DEBUG
@@ -283,6 +291,32 @@ void *PeriodicThread::thread_func( void * _data ) {
 
     vector< PeriodicThread::CallbackList::iterator > to_remove;
     thread->callback_lock.lock();
+
+    // At the beginning of every haptic thread cycle, add more callbacks if 
+    // there are any. So the added call back do not need to wait one cycle to get
+    // added , and at least another cycle to actually get executed in worst case.
+    // This will make the synchronousHapticCB runs faster as that function will wait
+    // for sync_haptics callback to be actually executed in haptic thread. However,
+    // even the callback adding in the very beginning of the cycle, in worst case, 
+    // the callback may still miss its first chance to get added 
+    // and wait one cycle more to be added and executed. So depends on how fast haptic
+    // thread runs and how many haptic thread in total, synchronousHapticCB may 
+    // becomes relatively slow.
+    
+    // if no more callbacks wait for a callback to be added in order to
+    // avoid spending time doing no useful operations in the thread.
+    thread->transferCallbackList();
+    if( thread->frequency < 0 ) {
+      // Check thread_func_is_running in order to avoid potential hangs
+      // if the user clears the callbacks list and then destroys the thread
+      // class. In that case the wait statement can be reached after
+      // signal in ~PeriodicThread().
+      if( thread->callbacks.size() == 0 && thread->thread_func_is_running ) {
+        thread->callback_lock.wait();
+        thread->transferCallbackList();
+      }
+    }
+
     for( PeriodicThread::CallbackList::iterator i = thread->callbacks.begin();
          i != thread->callbacks.end(); ++i ) {
       PeriodicThread::CallbackCode c = ( (*i).second ).first(
@@ -304,23 +338,10 @@ void *PeriodicThread::thread_func( void * _data ) {
 
     thread->callback_lock.signal();
 
-    // if no more callbacks wait for a callback to be added in order to
-    // avoid spending time doing no useful operations in the thread.
-    if( thread->frequency < 0 ) {
-      // Check thread_func_is_running in order to avoid potential hangs
-      // if the user clears the callbacks list and then destroys the thread
-      // class. In that case the wait statement can be reached after
-      // signal in ~PeriodicThread().
-      if( thread->callbacks.size() == 0 && thread->thread_func_is_running )
-        thread->callback_lock.wait();
-    } else {
-      thread->transferCallbackList();
-    }
-
     thread->callback_lock.unlock();
 
     sched_yield();
-#ifndef WIN32
+#ifndef H3D_WINDOWS
     // According to documentation usleep(0) should not do anything so it should
     // not do any harm to add this code. The reason to add it is that
     // sched_yield does not seem to do what we expect it to
@@ -330,8 +351,9 @@ void *PeriodicThread::thread_func( void * _data ) {
     // is used. Therefore usleep(0) is added because it must do something
     // since the lock is suddenly obtained.
     usleep(0);
-#endif
+#endif // H3D_WINDOWS
   }
+  return 0;
 } 
 
 ThreadBase::ThreadId ThreadBase::main_thread_id =
@@ -354,7 +376,7 @@ PeriodicThread::PeriodicThread( int _thread_priority,
                                 int _thread_frequency ):
   frequency( _thread_frequency ),
   thread_func_is_running( true ) {
-#ifdef WIN32
+#ifdef H3D_WINDOWS
   priority = _thread_priority == THREAD_PRIORITY_LOWEST ? LOW_PRIORITY :
              _thread_priority == THREAD_PRIORITY_NORMAL ? NORMAL_PRIORITY :
              _thread_priority == THREAD_PRIORITY_ABOVE_NORMAL ? HIGH_PRIORITY :
@@ -370,7 +392,7 @@ PeriodicThread::PeriodicThread( int _thread_priority,
   int policy =
     priority > NORMAL_PRIORITY ?
     SCHED_FIFO : SCHED_OTHER;
-#endif
+#endif //H3D_WINDOWS
 
   pthread_attr_t attr;
   sched_param p;
@@ -394,7 +416,7 @@ PeriodicThread::PeriodicThread( Priority _thread_priority,
   sched_param p;
   pthread_attr_getschedparam( &attr, &p );
   
-#ifdef WIN32
+#ifdef H3D_WINDOWS
   
   p.sched_priority =
     priority == LOW_PRIORITY    ? THREAD_PRIORITY_LOWEST :
@@ -422,7 +444,7 @@ PeriodicThread::PeriodicThread( Priority _thread_priority,
     priority > NORMAL_PRIORITY ?
     SCHED_FIFO : SCHED_OTHER;
   
-#endif
+#endif // H3D_WINDOWS
   
   pthread_attr_setschedparam( &attr, &p );
   pthread_attr_setschedpolicy( &attr, policy );
@@ -450,25 +472,13 @@ PeriodicThread::~PeriodicThread() {
   }
 }
 
-SimpleThread::SimpleThread( void *(func) (void *),
-                            void *args,
-                            int thread_priority ) {
-  pthread_attr_t attr;
-  sched_param p;
-  p.sched_priority = thread_priority;
-  pthread_attr_init( &attr );
-  pthread_attr_setschedparam( &attr, &p );
-  pthread_create( &thread_id, &attr, func, args ); 
-  pthread_attr_destroy( &attr );
-}
-
-SimpleThread::SimpleThread( void *(func) (void *),
+SimpleThread::SimpleThread( FunctionPtr func,
                             void *args,
                             Priority thread_priority ){
   pthread_attr_t attr;
   sched_param p;
   
-#ifdef WIN32
+#ifdef H3D_WINDOWS
   
   p.sched_priority =
     thread_priority == LOW_PRIORITY    ? THREAD_PRIORITY_LOWEST :
@@ -496,12 +506,12 @@ SimpleThread::SimpleThread( void *(func) (void *),
     thread_priority > NORMAL_PRIORITY ?
     SCHED_FIFO : SCHED_OTHER;
   
-#endif
+#endif // H3D_WINDOWS
 
   pthread_attr_init( &attr );
   pthread_attr_setschedparam( &attr, &p );
   pthread_attr_setschedpolicy( &attr, policy );
-  pthread_create( &thread_id, &attr, func, args ); 
+  valid = (pthread_create( &thread_id, &attr, func, args ) == 0);
   pthread_attr_destroy( &attr );
 }
 
@@ -510,6 +520,10 @@ int SimpleThread::join() {
 }
 
 SimpleThread::~SimpleThread() {
+  if( !valid ) {// if not valid, do not need to kill the thread
+    Console( LogLevel::Info ) << "no need to kill invalid thread: " << name << std::endl;
+    return;
+  }
   ThreadId this_thread = getCurrentThreadId();
   if( !pthread_equal( this_thread, thread_id ) ) {
     pthread_cancel( thread_id );  
@@ -524,7 +538,9 @@ void PeriodicThread::synchronousCallback( CallbackFunc func, void *data ) {
   // signal the thread that a new callback is available if it is waiting for one.
   if( frequency < 0 && callbacks.size() == 0 ) callback_lock.signal();
   // add the new callback.
+  callbacks_added_lock.lock();  // Have to get added_lock in order to call genCallbackId in a thread safe way
   callbacks.push_back( make_pair( genCallbackId(), make_pair( func, data ) ) );
+  callbacks_added_lock.unlock();
   // wait for the callback to be done.
   callback_lock.wait();
   callback_lock.unlock();
@@ -532,21 +548,22 @@ void PeriodicThread::synchronousCallback( CallbackFunc func, void *data ) {
 
 int PeriodicThread::asynchronousCallback( CallbackFunc func, void *data ) {
   int cb_id = -1;
-  if( frequency < 0 ) {
+  bool use_callback_lock = false;
+
+  callbacks_added_lock.lock();
+  cb_id = genCallbackId();
+  // add the new callback
+  callbacks_added.push_back( make_pair( cb_id, make_pair( func, data ) ) );
+  if( frequency < 0 && callbacks.size() == 0 )
+    use_callback_lock = true;
+  callbacks_added_lock.unlock();
+
+  if( use_callback_lock ) {
     callback_lock.lock();
     // signal the thread that a new callback is available if it is waiting for
     // one.
     if( callbacks.size() == 0 ) callback_lock.signal();
-    cb_id = genCallbackId();
-    // add the new callback
-    callbacks.push_back( make_pair( cb_id, make_pair( func, data ) ) );
     callback_lock.unlock();
-  } else {
-    callbacks_added_lock.lock();
-    cb_id = genCallbackId();
-    // add the new callback
-    callbacks_added.push_back( make_pair( cb_id, make_pair( func, data ) ) );
-    callbacks_added_lock.unlock();
   }
   return cb_id;
 }
@@ -637,12 +654,16 @@ HapticThreadBase::~HapticThreadBase() {
 
 bool HapticThreadBase::inHapticThread() {
   PeriodicThread::ThreadId id = PeriodicThread::getCurrentThreadId();
-  for( vector< HapticThreadBase *>::iterator i = threads.begin();
+  sg_lock.lock();
+  for( vector< HapticThreadBase *>::const_iterator i = threads.begin();
        i != threads.end(); ++i ) {
     ThreadBase *thread = dynamic_cast< ThreadBase * >( *i );
-    if( thread && pthread_equal( thread->getThreadId(), id ) )
+    if( thread && pthread_equal( thread->getThreadId(), id ) ) {
+      sg_lock.unlock();
       return true;
+    }
   }
+  sg_lock.unlock();
   return false;
 }
 
@@ -739,9 +760,11 @@ typedef struct tagTHREADNAME_INFO {
 #endif
 
 void ThreadBase::setThreadName( ThreadId thread_id, const string &name ) {
+#if ( (PTW32_VERSION_MAJOR >= 3)  || ( ( PTW32_VERSION_MAJOR == 2 ) && ( PTW32_VERSION_MINOR >= 11 ) ) )
+  pthread_setname_np( thread_id,  name.c_str() );
+#endif
 
 #ifdef _MSC_VER
- Sleep(10);
  THREADNAME_INFO info;
  info.dwType = 0x1000;
  info.szName = name.c_str();
@@ -755,7 +778,9 @@ void ThreadBase::setThreadName( ThreadId thread_id, const string &name ) {
  __except(EXCEPTION_EXECUTE_HANDLER)
  {
  }
- 
+#ifdef HAVE_NVIDIATX
+  nvtxNameOsThread(info.dwThreadID, name.c_str());
+#endif
 #endif
 
 }
@@ -773,18 +798,12 @@ bool ThreadBase::inMainThread() {
   return pthread_equal( main_thread_id, getCurrentThreadId() ) != 0;
 } 
 
+ThreadSpecificContainerBase::ThreadInfo::ThreadInfo() {
+  ScopedLock< MutexLock > lock( mutex );
+  index = next_thread_index++;
+}
 
-
-
-
-
-
-
-
-
-
-
-
-
-
-
+ThreadSpecificContainerBase::ThreadInfo& ThreadSpecificContainerBase::ThreadInfo::getInstance() {
+  static ThreadSpecificStorage< ThreadInfo > thread_info;
+  return thread_info;
+}
